#!/usr/bin/python3

########################################################################
#                                                                      #
# cubic.py                                                             #
#                                                                      #
# Copyright (C) 2015 PJ Singh <psingh.cubic@gmail.com>                 #
#                                                                      #
########################################################################

########################################################################
#                                                                      #
# This file is part of Cubic - Custom Ubuntu ISO Creator.              #
#                                                                      #
# Cubic is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by #
# the Free Software Foundation, either version 3 of the License, or    #
# (at your option) any later version.                                  #
#                                                                      #
# Cubic is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of       #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         #
# GNU General Public License for more details.                         #
#                                                                      #
# You should have received a copy of the GNU General Public License    #
# along with Cubic. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                      #
########################################################################

import logger
import display
from handlers import MainWindowHandlers
import model
import utilities

import gi
gi.require_version('Gdk', '3.0')
from gi.repository import Gdk
from gi.repository import GLib
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

import pwd
import os
import sys
import traceback

try:
    logger.log_step('Starting Cubic')

    root_uid = os.getuid()
    model.set_root_user_id(root_uid)
    root_gid = os.getgid()
    model.set_root_group_id(root_gid)

    working_directory = sys.argv[1]
    os.chdir(working_directory)
    current_directory = os.getcwd()
    logger.log_data('The working directory is', current_directory)

    uid = int(os.getenv('PKEXEC_UID'))
    pw_name, pw_passwd, pw_uid, pw_gid, pw_gecos, pw_dir, pw_shell = pwd.getpwuid(
        uid)

    model.set_user_name(pw_name)
    model.set_user_id(pw_uid)
    model.set_group_id(pw_gid)
    model.set_home(os.path.expanduser(pw_dir))
    os.environ['HOME'] = model.home
    logger.log_data('Set the HOME environment variable to', model.home)

    theme = Gtk.Settings.get_default().get_property('gtk-theme-name')
    logger.log_data('The current GTK theme is', theme)

    gtk_version = utilities.get_package_version('gir1.2-gtk-3.0')[0:4]
    logger.log_data('The GTK version is', gtk_version[0:4])
    if float(gtk_version) < 3.20:
        model.set_cubic_ui_filename('cubic_gtk310.ui')
    else:
        # model.set_cubic_ui_filename('cubic.ui')
        model.set_cubic_ui_filename('cubic_gtk320.ui')

    model.set_builder(Gtk.Builder())
    model.builder.add_from_file(model.cubic_ui_filename)

    # TODO: Temporary workaround for error:
    # (cubic.py:2790): dconf-CRITICAL **: 09:34:53.072: unable to create file
    # '/home/<user>/.cache/dconf/user': Permission denied. dconf will not work
    # properly.
    dconf_directory = os.path.join(model.home, '.cache', 'dconf')
    try:
        os.chown(dconf_directory, model.user_id, model.group_id)
    except FileNotFoundError:
        pass

    handlers = MainWindowHandlers()
    model.builder.connect_signals(handlers)

    model.set_page_name('project_directory_page')

    cubic_version = utilities.get_package_version('cubic')
    model.set_cubic_version(cubic_version)
    display.update_label(
        'project_directory_page__version_label',
        model.cubic_version[0:10])

    window = model.builder.get_object('window')

    css_provider = Gtk.CssProvider()

    if theme == 'Ambiance' or theme == 'Radiance':
        css_provider.load_from_path('cubic_unity.css')
    else:
        css_provider.load_from_path('cubic.css')

    screen = Gdk.Screen.get_default()
    style_context = window.get_style_context()
    style_context.add_provider_for_screen(
        screen,
        css_provider,
        Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    window.show()

    GLib.threads_init()
    Gdk.threads_enter()
    Gtk.main()
    Gdk.threads_leave()

except IndexError as exception:
    logger.log_step(
        'Error: Cubic (Custom Ubuntu ISO Creator) is a graphical user interface application and should be run using the application launcher. See "man cubic" for more information.'
    )
    logger.log_data('The tracekback is', traceback.format_exc())
except TypeError as exception:
    logger.log_step(
        'Error: Cubic (Custom Ubuntu ISO Creator) is a graphical user interface application and should be run using the application launcher. See "man cubic" for more information.'
    )
    logger.log_data('The tracekback is', traceback.format_exc())
except Exception as exception:
    logger.log_step(
        'Error: Cubic (Custom Ubuntu ISO Creator) is a graphical user interface application and should be run using the application launcher. See "man cubic" for more information.'
    )
    logger.log_data('The tracekback is', traceback.format_exc())
